package dev.fitko.fitconnect.core;

import com.fasterxml.jackson.core.JsonProcessingException;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.event.EventLog;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.event.EventLogPayload;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.submission.*;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.*;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.api.services.destination.DestinationService;
import dev.fitko.fitconnect.api.services.destination.KeyService;
import dev.fitko.fitconnect.api.services.events.EventLogService;
import dev.fitko.fitconnect.api.services.events.EventLogVerificationService;
import dev.fitko.fitconnect.api.services.submission.SubmissionService;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import dev.fitko.fitconnect.core.util.EventLogUtil;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import java.util.List;
import java.util.UUID;
import java.util.stream.Collectors;

public class SubmissionSender implements Sender {

    private static final Logger LOGGER = LoggerFactory.getLogger(SubmissionSender.class);

    private static final ObjectMapper MAPPER = new ObjectMapper();

    private final ValidationService validationService;
    private final CryptoService cryptoService;
    private final SubmissionService submissionService;
    private final EventLogService eventLogService;
    private final EventLogVerificationService eventLogVerifier;
    private final KeyService keyService;
    private final DestinationService destinationService;

    public SubmissionSender(final SubmissionService submissionService,
                            final DestinationService destinationService,
                            final EventLogService eventLogService,
                            final EventLogVerificationService eventLogVerifier,
                            final CryptoService encryptionService,
                            final ValidationService validationService,
                            final KeyService keyService) {

        this.submissionService = submissionService;
        this.destinationService = destinationService;
        this.eventLogService = eventLogService;
        this.eventLogVerifier = eventLogVerifier;
        this.cryptoService = encryptionService;
        this.validationService = validationService;
        this.keyService = keyService;
    }

    @Override
    public ValidationResult validatePublicKey(final RSAKey publicKey) {
        LOGGER.info("Validating public key integrity");
        return this.validationService.validatePublicKey(publicKey);
    }

    @Override
    public ValidationResult validateMetadata(final Metadata metadata) {
        return this.validationService.validateMetadataSchema(metadata);
    }

    @Override
    public String encryptBytes(final RSAKey publicKey, final byte[] data) {
        LOGGER.info("Encrypting {} bytes", data.length);
        return this.cryptoService.encryptBytes(publicKey, data);
    }

    @Override
    public String createHash(final byte[] data) {
        LOGGER.info("Creating hash of {} bytes", data.length);
        return this.cryptoService.hashBytes(data);
    }

    @Override
    public SubmissionForPickup createSubmission(final CreateSubmission submission) {
        LOGGER.info("Creating new submission for destination {}", submission.getDestinationId());
        return this.submissionService.announceSubmission(submission);
    }

    @Override
    public void uploadAttachment(final UUID submissionId, final UUID attachmentId, final String encryptedAttachment) {
        LOGGER.info("Uploading attachment {} for submission {}", attachmentId, submissionId);
        this.submissionService.uploadAttachment(submissionId, attachmentId, encryptedAttachment);
    }

    @Override
    public Destination getDestination(final UUID destinationId) {
        return this.destinationService.getDestination(destinationId);
    }

    @Override
    public RSAKey getEncryptionKeyForDestination(final UUID destinationId) {
        LOGGER.info("Getting encryption key for destination id {}", destinationId);
        try {
            return this.keyService.getPublicEncryptionKey(destinationId);
        } catch (final RestApiException | InvalidKeyException e) {
            throw new KeyNotRetrievedException("Retrieving encryption JWK failed", e);
        }
    }

    @Override
    public String encryptObject(final RSAKey encryptionKey, final Object obj) {
        try {
            return this.cryptoService.encryptString(encryptionKey, MAPPER.writeValueAsString(obj));
        } catch (final JsonProcessingException | EncryptionException e) {
            throw new EncryptionException("Failure encrypting object", e);
        }
    }

    @Override
    public Submission sendSubmission(final SubmitSubmission submission) {
        LOGGER.info("Sending submission {}", submission.getSubmissionId());
        return this.submissionService.sendSubmission(submission);
    }

    @Override
    public ValidationResult validateJsonFormat(final String json) {
        return this.validationService.validateJsonFormat(json);
    }

    @Override
    public ValidationResult validateXmlFormat(final String xml) {
        return this.validationService.validateXmlFormat(xml);
    }

    @Override
    public List<EventLogEntry> getEventLog(final EventLogPayload eventLogPayload) {
        final EventLog eventLog = this.eventLogService.getEventLog(eventLogPayload.getCaseId());
        final List<ValidationResult> results = this.eventLogVerifier.validateEventLog(eventLogPayload, eventLog);
        if (!results.isEmpty()) {
            final String errorMessages = results.stream()
                    .map(ValidationResult::getError)
                    .map(Throwable::getMessage)
                    .distinct()
                    .collect(Collectors.joining("\n"));
            throw new EventLogException("Event log contains invalid entries: \n" + errorMessages);
        }
        return EventLogUtil.mapEventLogToEntries(eventLog);
    }
}
