package dev.fitko.fitconnect.core.util;

import com.nimbusds.jwt.JWTClaimsSet;
import com.nimbusds.jwt.SignedJWT;
import dev.fitko.fitconnect.api.domain.model.event.*;
import dev.fitko.fitconnect.api.domain.model.event.problems.Problem;
import dev.fitko.fitconnect.api.exceptions.EventLogException;

import java.text.ParseException;
import java.util.ArrayList;
import java.util.List;
import java.util.Map;
import java.util.UUID;
import java.util.stream.Collectors;

import static dev.fitko.fitconnect.api.domain.model.event.EventIssuer.*;

public final class EventLogUtil {

    public static List<EventLogEntry> mapEventLogToEntries(final EventLog eventLog) {
        return eventLog.getEventLogs()
                .stream()
                .map(EventLogUtil::eventToEntry)
                .collect(Collectors.toList());
    }

    public static EventIssuer resolveIssuerType(final String issuer) {
        if(issuer == null){
            return UNDEFINED;
        }
        return issuer.startsWith("http") ? SUBMISSION_SERVICE : DESTINATION;
    }

    public static Event getEventTypeFromEvent(final JWTClaimsSet claims) {
        final Map<String, Object> events = (Map) claims.getClaim(EventClaimFields.CLAIM_EVENTS);
        final Map.Entry<String, Map> eventEntry = (Map.Entry) events.entrySet().iterator().next();
        return Event.fromString(eventEntry.getKey());
    }

    public static UUID getDestinationId(final String issuer) {
        final EventIssuer issuerType = resolveIssuerType(issuer);
        if (issuerType.equals(SUBMISSION_SERVICE) || issuerType.equals(UNDEFINED)) {
            return null;
        }
        try {
            return UUID.fromString(issuer);
        } catch (final IllegalArgumentException ex) {
            throw new EventLogException("Destination id '" + issuer + "' from issuer is no valid uuid");
        }
    }

    private static EventLogEntry eventToEntry(final String setEvent)  {

        final SignedJWT signedJWT = parseEvent(setEvent);
        final JWTClaimsSet claimsSet = getClaimsSet(signedJWT);
        final Map<String, Object> events = (Map) claimsSet.getClaim(EventClaimFields.CLAIM_EVENTS);
        final Map.Entry<String, Map> eventEntry = (Map.Entry) events.entrySet().iterator().next();
        final String eventName = eventEntry.getKey();
        final Map<String, Object> eventPayload = eventEntry.getValue();
        final List<Problem> problems = new ArrayList<>();
        if(eventPayload.containsKey("problems")){
            problems.addAll((List<Problem>)eventPayload.get("problems"));
        }
        return EventLogEntry.builder()
                .event(eventName)
                .issuer(claimsSet.getIssuer())
                .issueTime(claimsSet.getIssueTime())
                .eventId(getEventId(claimsSet))
                .caseId(getTransactionClaim(claimsSet))
                .submissionId(getSubmissionId(claimsSet))
                .problems(problems)
                .build();
    }

    private static UUID getSubmissionId(final JWTClaimsSet claimsSet) {
        return TypeAndUUID.fromString(claimsSet.getSubject()).getUuid();
    }

    private static UUID getEventId(final JWTClaimsSet claimsSet) {
        return UUID.fromString(claimsSet.getJWTID());
    }

    private static UUID getTransactionClaim(final JWTClaimsSet claimsSet)  {
        try {
            return TypeAndUUID.fromString(claimsSet.getStringClaim(EventClaimFields.CLAIM_TXN)).getUuid();
        } catch (final ParseException e) {
            throw new EventLogException(e.getMessage(), e);
        }
    }

    private static JWTClaimsSet getClaimsSet(final SignedJWT signedJWT) {
        try {
            return signedJWT.getJWTClaimsSet();
        } catch (final ParseException e) {
            throw new EventLogException(e.getMessage(), e);
        }
    }

    private static SignedJWT parseEvent(final String event) {
        try {
            return SignedJWT.parse(event);
        } catch (final ParseException e) {
            throw new EventLogException(e.getMessage(), e);
        }
    }

}
