package dev.fitko.fitconnect.core;

import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.jwk.KeyUse;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jose.jwk.gen.RSAKeyGenerator;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.event.EventLog;
import dev.fitko.fitconnect.api.domain.model.event.EventLogEntry;
import dev.fitko.fitconnect.api.domain.model.event.EventLogPayload;
import dev.fitko.fitconnect.api.domain.model.metadata.Metadata;
import dev.fitko.fitconnect.api.domain.model.submission.*;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.*;
import dev.fitko.fitconnect.api.services.Sender;
import dev.fitko.fitconnect.api.services.crypto.CryptoService;
import dev.fitko.fitconnect.api.services.destination.DestinationService;
import dev.fitko.fitconnect.api.services.destination.KeyService;
import dev.fitko.fitconnect.api.services.events.EventLogService;
import dev.fitko.fitconnect.api.services.events.EventLogVerificationService;
import dev.fitko.fitconnect.api.services.submission.SubmissionService;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.text.ParseException;
import java.util.Collections;
import java.util.List;
import java.util.UUID;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.*;
import static org.junit.jupiter.api.Assertions.*;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.*;

public class SubmissionSenderTest {

    private DestinationService destinationServiceMock;
    private SubmissionService submissionServiceMock;
    private EventLogService eventLogServiceMock;
    private EventLogVerificationService eventLogVerificationServiceMock;
    private CryptoService cryptoServiceMock;
    private ValidationService validationServiceMock;
    private KeyService keyServiceMock;
    private Sender underTest;

    @BeforeEach
    public void setUp() {
        this.destinationServiceMock = mock(DestinationService.class);
        this.submissionServiceMock = mock(SubmissionService.class);
        this.eventLogServiceMock = mock(EventLogService.class);
        this.cryptoServiceMock = mock(CryptoService.class);
        this.validationServiceMock = mock(ValidationService.class);
        this.eventLogVerificationServiceMock = mock(EventLogVerificationService.class);
        this.keyServiceMock = mock(KeyService.class);

        this.underTest = new SubmissionSender(
                this.submissionServiceMock,
                this.destinationServiceMock,
                this.eventLogServiceMock,
                this.eventLogVerificationServiceMock,
                this.cryptoServiceMock,
                this.validationServiceMock,
                this.keyServiceMock
        );
    }

    @Test
    void validatePublicKey() throws JOSEException {

        //Given
        final RSAKey publicKey = new RSAKeyGenerator(2048)
                .keyUse(KeyUse.ENCRYPTION)
                .keyID(UUID.randomUUID().toString())
                .generate()
                .toPublicJWK();

        when(this.validationServiceMock.validatePublicKey(any())).thenReturn(ValidationResult.ok());

        // When
        final ValidationResult validationResult = this.underTest.validatePublicKey(publicKey);

        // Then
        assertTrue(validationResult.isValid());
    }

    @Test
    void encryptBytes() throws JOSEException {

        //Given
        final RSAKey publicKey = new RSAKeyGenerator(2048)
                .keyUse(KeyUse.ENCRYPTION)
                .keyID(UUID.randomUUID().toString())
                .generate()
                .toPublicJWK();

        when(this.cryptoServiceMock.encryptBytes(any(), any())).thenReturn("encryptedString");

        // When
        final String encryptedData = this.underTest.encryptBytes(publicKey, "data".getBytes());

        // Then
        assertNotNull(encryptedData);
        assertThat(encryptedData, is("encryptedString"));
    }

    @Test
    void createHash() {

        // Given
        when(this.cryptoServiceMock.hashBytes(any())).thenReturn("hashedData");

        // When
        final String hashedData = this.underTest.createHash("data".getBytes());

        // Then
        assertNotNull(hashedData);
        assertThat(hashedData, is("hashedData"));
    }

    @Test
    void createSubmission() {

        // Given
        final var createSubmission = CreateSubmission.builder()
                .caseId(UUID.randomUUID())
                .destinationId(UUID.randomUUID())
                .announcedAttachments(Collections.emptyList())
                .serviceType(new ServiceType())
                .build();

        final var expectedSubmission = new SubmissionForPickup();
        expectedSubmission.setSubmissionId(UUID.randomUUID());
        expectedSubmission.setDestinationId(createSubmission.getDestinationId());
        expectedSubmission.setCaseId(createSubmission.getCaseId());

        when(this.submissionServiceMock.announceSubmission(any())).thenReturn(expectedSubmission);

        // When
        final SubmissionForPickup submissionForPickup = this.underTest.createSubmission(createSubmission);

        // Then
        assertThat(submissionForPickup, equalTo(expectedSubmission));
    }

    @Test
    void uploadAttachment() {

        // Given
        final var submissionId = UUID.randomUUID();
        final var attachmentId = UUID.randomUUID();
        final var encryptedAttachment = "encryptedAttachmentBytes";

        doNothing().when(this.submissionServiceMock).uploadAttachment(any(), any(), any());

        // When
        this.underTest.uploadAttachment(submissionId, attachmentId, encryptedAttachment);

        // Then
        verify(this.submissionServiceMock, times(1)).uploadAttachment(submissionId, attachmentId, encryptedAttachment);
    }

    @Test
    void getEncryptionKeyForDestination() throws ParseException, IOException {

        // Given
        final var submissionId = UUID.randomUUID();

        final String publicKeyJson = getResourceAsString("/public_encryption_test_key.json");
        final RSAKey expectedRSAKey = RSAKey.parse(publicKeyJson);

        when(this.keyServiceMock.getPublicEncryptionKey(any())).thenReturn(expectedRSAKey);

        // When
        final RSAKey key = this.underTest.getEncryptionKeyForDestination(submissionId);

        // Then
        assertThat(key, equalTo(expectedRSAKey));
    }

    @Test
    void getEncryptionKeyForDestinationFailed() {

        // Given
        when(this.keyServiceMock.getPublicEncryptionKey(any())).thenThrow(RestApiException.class);

        // Then
        assertThrows(
                KeyNotRetrievedException.class,
                () -> this.underTest.getEncryptionKeyForDestination(UUID.randomUUID()),
                "Expected getEncryptionKeyForDestination() to throw, but nothing was thrown"
        );
    }

    @Test
    void encryptObject() throws JOSEException {

        //Given
        final RSAKey publicKey = new RSAKeyGenerator(2048)
                .keyUse(KeyUse.ENCRYPTION)
                .keyID(UUID.randomUUID().toString())
                .generate()
                .toPublicJWK();

        when(this.cryptoServiceMock.encryptString(any(), any())).thenReturn("encryptedObject");

        final Metadata metadata = new Metadata();

        // When
        final String encryptedMetadata = this.underTest.encryptObject(publicKey, metadata);

        // Then
        assertThat(encryptedMetadata, equalTo("encryptedObject"));
    }

    @Test
    void encryptingObjectFailed() throws JOSEException {

        //Given
        final RSAKey publicKey = new RSAKeyGenerator(2048)
                .keyUse(KeyUse.ENCRYPTION)
                .keyID(UUID.randomUUID().toString())
                .generate()
                .toPublicJWK();

        // When
        when(this.cryptoServiceMock.encryptString(any(), any())).thenThrow(EncryptionException.class);

        // Then
        assertThrows(
                EncryptionException.class,
                () -> this.underTest.encryptObject(publicKey, new Metadata()),
                "Expected encryptObject() to throw, but nothing was thrown"
        );
    }

    @Test
    void sendSubmission() {

        // Given
        final SubmitSubmission submitSubmission = new SubmitSubmission();
        submitSubmission.setSubmissionId(UUID.randomUUID());
        submitSubmission.setEncryptedData("encryptedData");
        submitSubmission.setEncryptedMetadata("encryptedMetadata");

        final Submission expectedSubmission = new Submission();
        expectedSubmission.setSubmissionId(submitSubmission.getSubmissionId());
        expectedSubmission.setEncryptedData(submitSubmission.getEncryptedData());
        expectedSubmission.setEncryptedMetadata(submitSubmission.getEncryptedMetadata());

        when(this.submissionServiceMock.sendSubmission(any())).thenReturn(expectedSubmission);

        // When
        this.underTest.sendSubmission(submitSubmission);

        // Then
        verify(this.submissionServiceMock, times(1)).sendSubmission(submitSubmission);
    }

    @Test
    void getDestination() {

        // Given
        final var destinationId = UUID.randomUUID();

        final var expectedDestination = new Destination();
        expectedDestination.setDestinationId(destinationId);

        when(this.destinationServiceMock.getDestination(any())).thenReturn(expectedDestination);

        // When
        final Destination destination = this.underTest.getDestination(destinationId);

        // Then
        assertThat(destination, equalTo(expectedDestination));
    }

    @Test
    void testValidMetadata() {

        // Given
        final var metadata = new Metadata();

        when(this.validationServiceMock.validateMetadataSchema(any())).thenReturn(ValidationResult.ok());

        // When
        final ValidationResult validatedMetadata = this.underTest.validateMetadata(metadata);

        // Then
        assertTrue(validatedMetadata.isValid());
    }

    @Test
    void testInvalidMetadata() {

        // Given
        final var invalidMetadata = new Metadata();
        final ValidationResult expectedResult = ValidationResult.error(new ValidationException("invalid metadata"));
        when(this.validationServiceMock.validateMetadataSchema(any())).thenReturn(expectedResult);

        // When
        final ValidationResult validatedMetadata = this.underTest.validateMetadata(invalidMetadata);

        // Then
        assertTrue(validatedMetadata.hasError());
        assertThat(validatedMetadata.getError().getMessage(), equalTo("invalid metadata"));
    }

    @Test
    void testValidJsonFormatValidation(){

        // Given
        when(this.validationServiceMock.validateJsonFormat(any())).thenReturn(ValidationResult.ok());

        // When
        final ValidationResult validationResult = this.underTest.validateJsonFormat("{}");

        //Then
        assertTrue(validationResult.isValid());
    }

    @Test
    void testInvalidJsonFormatValidation(){

        // Given
        when(this.validationServiceMock.validateJsonFormat(any())).thenReturn(ValidationResult.error(new ValidationException("Invalid json")));

        // When
        final ValidationResult validationResult = this.underTest.validateJsonFormat("$%&/()");

        //Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("Invalid json"));
    }

    @Test
    void testValidXmlFormatValidation(){

        // Given
        when(this.validationServiceMock.validateXmlFormat(any())).thenReturn(ValidationResult.ok());

        // When
        final ValidationResult validationResult = this.underTest.validateXmlFormat("<xml>test</xml>");

        //Then
        assertTrue(validationResult.isValid());
    }

    @Test
    void testInvalidXmlFormatValidation(){

        // Given
        when(this.validationServiceMock.validateXmlFormat(any())).thenReturn(ValidationResult.error(new ValidationException("Invalid xml")));

        // When
        final ValidationResult validationResult = this.underTest.validateXmlFormat("$%&/(");

        //Then
        assertTrue(validationResult.hasError());
        assertThat(validationResult.getError().getMessage(), containsString("Invalid xml"));
    }

    @Test
    void testReceiveEventLogWithValidValidation() {

        // Given
        final EventLog emptyLog = new EventLog();

        when(this.eventLogServiceMock.getEventLog(any())).thenReturn(emptyLog);
        when(this.validationServiceMock.validateSetEventSchema(any())).thenReturn(ValidationResult.ok());

        // When
        final List<EventLogEntry> eventLog = this.underTest.getEventLog(new EventLogPayload());

        // Then
        assertTrue(eventLog.isEmpty());
    }

    @Test
    void testReceiveEventLogWithInvalidValidation() {

        // Given
        final EventLog emptyLog = new EventLog();
        final ValidationResult error = ValidationResult.error(new ValidationException("Invalid log"));

        when(this.eventLogServiceMock.getEventLog(any())).thenReturn(emptyLog);
        when(this.eventLogVerificationServiceMock.validateEventLog(any(), any())).thenReturn(List.of(error));

        // Then
        assertThrows(EventLogException.class, () ->  this.underTest.getEventLog(new EventLogPayload()));
    }


    private String getResourceAsString(final String filename) throws IOException {
        return new String(SubmissionSenderTest.class.getResourceAsStream(filename).readAllBytes());
    }
}