package dev.fitko.fitconnect.core.destination;

import com.fasterxml.jackson.databind.ObjectMapper;
import com.nimbusds.jose.jwk.RSAKey;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.config.Environment;
import dev.fitko.fitconnect.api.config.EnvironmentName;
import dev.fitko.fitconnect.api.domain.model.destination.Destination;
import dev.fitko.fitconnect.api.domain.model.jwk.ApiJwk;
import dev.fitko.fitconnect.api.domain.model.jwk.ApiJwkSet;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.InvalidKeyException;
import dev.fitko.fitconnect.api.services.destination.DestinationService;
import dev.fitko.fitconnect.api.services.destination.KeyService;
import dev.fitko.fitconnect.api.services.portal.SelfServicePortalService;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.IOException;
import java.text.ParseException;
import java.util.List;
import java.util.Map;
import java.util.UUID;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.assertThrows;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

class PublicKeyServiceTest {

    private KeyService underTest;

    private DestinationService destinationServiceMock;

    private SelfServicePortalService portalServiceMock;

    private ValidationService validationServiceMock;

    @BeforeEach
    void setup() {
        this.destinationServiceMock = mock(DestinationService.class);
        this.portalServiceMock = mock(SelfServicePortalService.class);
        this.validationServiceMock = mock(ValidationService.class);
        this.underTest = getKeyService(true);
    }

    @Test
    void getPublicEncryptionKey() throws IOException, ParseException {

        // Given
        final Destination destination = new Destination();
        destination.setDestinationId(UUID.randomUUID());
        destination.setEncryptionKid("123");

        final String publicKeyJson = getResourceAsString("/public_encryption_test_key.json");

        final ApiJwk jwk = new ObjectMapper().readValue(publicKeyJson, ApiJwk.class);

        when(this.destinationServiceMock.getDestination(any())).thenReturn(destination);
        when(this.destinationServiceMock.getPublicKey(any(), any())).thenReturn(jwk);
        when(this.validationServiceMock.validatePublicKey(any())).thenReturn(ValidationResult.ok());

        // When
        final RSAKey rsaKey = this.underTest.getPublicEncryptionKey(UUID.randomUUID());

        // Then
        assertThat(RSAKey.parse(publicKeyJson), is(rsaKey));
    }

    @Test
    void getPublicSignatureKey() throws IOException, ParseException {

        // Given
        final String publicKeyJson = getResourceAsString("/public_encryption_test_key.json");
        final ApiJwk jwk = new ObjectMapper().readValue(publicKeyJson, ApiJwk.class);
        final UUID destinationId = UUID.randomUUID();

        when(this.destinationServiceMock.getPublicKey(destinationId, jwk.getKid())).thenReturn(jwk);
        when(this.validationServiceMock.validatePublicKey(any())).thenReturn(ValidationResult.ok());

        // When
        final RSAKey rsaKey = this.underTest.getPublicSignatureKey(destinationId, jwk.getKid());

        // Then
        assertThat(RSAKey.parse(publicKeyJson), is(rsaKey));
    }

    @Test
    void getPortalSignatureKey() throws IOException, ParseException {

        // Given
        final String publicKeyJson = getResourceAsString("/public_encryption_test_key.json");
        final ApiJwk jwk = new ObjectMapper().readValue(publicKeyJson, ApiJwk.class);
        final UUID destinationId = UUID.randomUUID();

        when(this.destinationServiceMock.getPublicKey(destinationId, jwk.getKid())).thenReturn(jwk);
        when(this.validationServiceMock.validatePublicKey(any())).thenReturn(ValidationResult.ok());
        when(this.portalServiceMock.getPublicKeys()).thenReturn(new ApiJwkSet(List.of(jwk)));

        // When
        final RSAKey rsaKey = this.underTest.getPortalSignatureKey(jwk.getKid());

        // Then
        assertThat(RSAKey.parse(publicKeyJson), is(rsaKey));
    }

    @Test
    void testKeyValidationIfInsecureKeyIsNotAllowed() throws IOException {

        // Given
        final PublicKeyService underTest = getKeyService(false);

        final String publicKeyJson = getResourceAsString("/public_encryption_test_key.json");
        final ApiJwk jwk = new ObjectMapper().readValue(publicKeyJson, ApiJwk.class);
        final UUID destinationId = UUID.randomUUID();

        when(this.destinationServiceMock.getPublicKey(destinationId, jwk.getKid())).thenReturn(jwk);
        when(this.validationServiceMock.validatePublicKey(any())).thenReturn(ValidationResult.error(new InvalidKeyException("Invalid key")));
        when(this.portalServiceMock.getPublicKeys()).thenReturn(new ApiJwkSet(List.of(jwk)));

        assertThrows(InvalidKeyException.class, () -> underTest.getPortalSignatureKey(jwk.getKid()));
    }

    private String getResourceAsString(final String filename) throws IOException {
        return new String(PublicKeyServiceTest.class.getResourceAsStream(filename).readAllBytes());
    }

    private PublicKeyService getKeyService(final boolean allowInsecureKey) {
        final EnvironmentName envName = new EnvironmentName("TESTING");
        final Environment environment = new Environment();
        environment.setAllowInsecurePublicKey(allowInsecureKey);

        final ApplicationConfig config = new ApplicationConfig();
        config.setEnvironments(Map.of(envName, environment));
        config.setActiveEnvironment(envName);

        return new PublicKeyService(config, this.destinationServiceMock, this.portalServiceMock, this.validationServiceMock);
    }
}