package dev.fitko.fitconnect.core.events;

import com.nimbusds.jose.JOSEException;
import com.nimbusds.jose.JWSHeader;
import com.nimbusds.jose.crypto.RSASSASigner;
import com.nimbusds.jose.jwk.RSAKey;
import com.nimbusds.jwt.JWTClaimsSet;
import com.nimbusds.jwt.SignedJWT;
import dev.fitko.fitconnect.api.config.ApplicationConfig;
import dev.fitko.fitconnect.api.domain.model.event.*;
import dev.fitko.fitconnect.api.domain.model.event.problems.Problem;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import dev.fitko.fitconnect.api.exceptions.SchemaNotFoundException;
import dev.fitko.fitconnect.api.services.destination.KeyService;
import dev.fitko.fitconnect.api.services.events.EventLogVerificationService;
import dev.fitko.fitconnect.api.services.validation.ValidationService;
import org.junit.jupiter.api.BeforeEach;
import org.junit.jupiter.api.Test;

import java.io.File;
import java.io.IOException;
import java.nio.file.Files;
import java.text.ParseException;
import java.util.*;

import static dev.fitko.fitconnect.api.domain.model.event.Event.ACCEPT;
import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.containsString;
import static org.junit.jupiter.api.Assertions.assertFalse;
import static org.junit.jupiter.api.Assertions.assertTrue;
import static org.mockito.ArgumentMatchers.any;
import static org.mockito.Mockito.mock;
import static org.mockito.Mockito.when;

class EventLogVerifierTest {

    private KeyService keyServiceMock;

    private ValidationService validationServiceMock;

    private EventLogVerificationService underTest;

    @BeforeEach
    void setup() {
        this.keyServiceMock = mock(KeyService.class);
        this.validationServiceMock = mock(ValidationService.class);
        this.underTest = new EventLogVerifier(this.keyServiceMock, this.validationServiceMock);
    }

    @Test
    void testValidLog() throws ParseException, JOSEException, IOException {

        // Given

        final RSAKey rsaKey = getRsaKey();
        final JWSHeader header = getDefaultJwsHeader(rsaKey);
        final Map<String, Map<String, Object>> events = getEvents(List.of(), ACCEPT);

        final UUID caseId = UUID.randomUUID();
        final UUID destinationId = UUID.randomUUID();
        final UUID submissionId = UUID.randomUUID();

        final JWTClaimsSet claims = getDefaultClaims(submissionId, caseId, destinationId, events);

        final String signedSET = getSignedSET(rsaKey, header, claims);

        final EventLog log = new EventLog();
        log.setEventLogs(List.of(signedSET));

        final EventLogPayload eventLogPayload = EventLogPayload.builder()
                .submissionId(submissionId)
                .destinationId(destinationId)
                .caseId(caseId)
                .build();

        when(this.validationServiceMock.validateSetEventSchema(any())).thenReturn(ValidationResult.ok());
        when(this.validationServiceMock.validateSignatureKey(any())).thenReturn(ValidationResult.ok());
        when(this.keyServiceMock.getPublicSignatureKey(any(),any())).thenReturn(rsaKey);

        // When
        final List<ValidationResult> validationResults = this.underTest.validateEventLog(eventLogPayload, log);

        // Then
        assertTrue(validationResults.isEmpty());
    }


    @Test
    void testMissingSchemaClaim() throws ParseException, JOSEException, IOException {

        // Given
        final RSAKey rsaKey = getRsaKey();
        final JWSHeader header = getDefaultJwsHeader(rsaKey);
        final Map<String, Map<String, Object>> events = getEvents(List.of(), ACCEPT);

        final UUID caseId = UUID.randomUUID();
        final UUID destinationId = UUID.randomUUID();
        final UUID submissionId = UUID.randomUUID();


        final JWTClaimsSet claims = new JWTClaimsSet.Builder()
                .issuer(destinationId.toString())
                .issueTime(new Date())
                .jwtID(UUID.randomUUID().toString())
                .subject("submission:" + submissionId)
                .claim(EventClaimFields.CLAIM_TXN, "case:" + caseId)
                .claim(EventClaimFields.CLAIM_EVENTS, events)
                .build();

        final String signedSET = getSignedSET(rsaKey, header, claims);

        final EventLog log = new EventLog();
        log.setEventLogs(List.of(signedSET));

        final EventLogPayload eventLogPayload = EventLogPayload.builder()
                .submissionId(submissionId)
                .destinationId(destinationId)
                .caseId(caseId)
                .build();

        when(this.validationServiceMock.validateSetEventSchema(any())).thenReturn(ValidationResult.ok());
        when(this.validationServiceMock.validateSignatureKey(any())).thenReturn(ValidationResult.ok());
        when(this.keyServiceMock.getPublicSignatureKey(any(),any())).thenReturn(rsaKey);

        // When
        final List<ValidationResult> validationResults = this.underTest.validateEventLog(eventLogPayload, log);

        // Then
        assertFalse(validationResults.isEmpty());
        assertThat(validationResults.get(0).getError().getMessage(), containsString("Referenced schema must not be null"));
    }

    @Test
    void testInvalidSchemaClaim() throws ParseException, JOSEException, IOException {

        // Given
        final RSAKey rsaKey = getRsaKey();
        final JWSHeader header = getDefaultJwsHeader(rsaKey);
        final Map<String, Map<String, Object>> events = getEvents(List.of(), ACCEPT);

        final UUID caseId = UUID.randomUUID();
        final UUID destinationId = UUID.randomUUID();
        final UUID submissionId = UUID.randomUUID();

        final JWTClaimsSet claims = new JWTClaimsSet.Builder()
                .claim(EventClaimFields.CLAIM_SCHEMA, "https://invalid-schema.fitko.de")
                .issuer(destinationId.toString())
                .issueTime(new Date())
                .jwtID(UUID.randomUUID().toString())
                .subject("submission:" + submissionId)
                .claim(EventClaimFields.CLAIM_TXN, "case:" + caseId)
                .claim(EventClaimFields.CLAIM_EVENTS, events)
                .build();

        final String signedSET = getSignedSET(rsaKey, header, claims);

        final EventLog log = new EventLog();
        log.setEventLogs(List.of(signedSET));

        final EventLogPayload eventLogPayload = EventLogPayload.builder()
                .submissionId(submissionId)
                .destinationId(destinationId)
                .caseId(caseId)
                .build();

        when(this.validationServiceMock.validateSetEventSchema(any())).thenReturn(ValidationResult.error(new SchemaNotFoundException("Schema not found")));
        when(this.validationServiceMock.validateSignatureKey(any())).thenReturn(ValidationResult.ok());
        when(this.keyServiceMock.getPublicSignatureKey(any(),any())).thenReturn(rsaKey);

        // When
        final List<ValidationResult> validationResults = this.underTest.validateEventLog(eventLogPayload, log);

        // Then
        assertFalse(validationResults.isEmpty());
        assertThat(validationResults.get(0).getError().getMessage(), containsString("Schema not found"));
    }

    @Test
    void testInvalidEventHeader() throws ParseException, JOSEException, IOException {

        // Given

        final JWSHeader header = JWSHeader.parse(Map.of(
                EventHeaderFields.TYPE, "WRONG_TYPE",
                EventHeaderFields.KEY_ID, "1234567",
                EventHeaderFields.ALGORITHM, "PS512",
                "ILLEGAL_4TH_FIELD", ""
        ));

        final RSAKey rsaKey = getRsaKey();
        final UUID caseId = UUID.randomUUID();
        final UUID destinationId = UUID.randomUUID();
        final UUID submissionId = UUID.randomUUID();

        final Map<String, Map<String, Object>> events = getEvents(List.of(), ACCEPT);
        final JWTClaimsSet claims = getDefaultClaims(submissionId, caseId, destinationId, events);
        final String signedSET = getSignedSET(rsaKey, header, claims);

        final EventLog log = new EventLog();
        log.setEventLogs(List.of(signedSET));

        final EventLogPayload eventLogPayload = EventLogPayload.builder()
                .submissionId(submissionId)
                .destinationId(destinationId)
                .caseId(caseId)
                .build();

        when(this.validationServiceMock.validateSetEventSchema(any())).thenReturn(ValidationResult.ok());
        when(this.validationServiceMock.validateSignatureKey(any())).thenReturn(ValidationResult.ok());
        when(this.keyServiceMock.getPublicSignatureKey(any(),any())).thenReturn(rsaKey);

        // When
        final List<ValidationResult> validationResults = this.underTest.validateEventLog(eventLogPayload, log);

        // Then
        assertFalse(validationResults.isEmpty());
        assertThat(validationResults.get(0).getError().getMessage(), containsString("The provided typ in the SET header is not secevent+jwt"));
        assertThat(validationResults.get(1).getError().getMessage(), containsString("Invalid keyId path parameter. Must be >=8 and <=64."));

    }

    @Test
    void testInvalidIssuerUUID() throws ParseException, JOSEException, IOException {

        // Given
        final RSAKey rsaKey = getRsaKey();
        final JWSHeader header = getDefaultJwsHeader(rsaKey);

        final UUID caseId = UUID.randomUUID();
        final UUID destinationId = UUID.randomUUID();
        final UUID submissionId = UUID.randomUUID();

        final Map<String, Map<String, Object>> events = getEvents(List.of(), ACCEPT);

        final JWTClaimsSet claims = new JWTClaimsSet.Builder()
                .claim(EventClaimFields.CLAIM_SCHEMA, new ApplicationConfig().getSetSchemaWriteVersion())
                .issuer("44567b87-6f6a---4f86-9d55-5ff98775d76f")
                .issueTime(new Date())
                .jwtID(UUID.randomUUID().toString())
                .subject("submission:" + submissionId)
                .claim(EventClaimFields.CLAIM_TXN, "case:" + caseId)
                .claim(EventClaimFields.CLAIM_EVENTS, events)
                .build();
        final String signedSET = getSignedSET(rsaKey, header, claims);

        final EventLog log = new EventLog();
        log.setEventLogs(List.of(signedSET));

        final EventLogPayload eventLogPayload = EventLogPayload.builder()
                .submissionId(submissionId)
                .destinationId(destinationId)
                .caseId(caseId)
                .build();

        when(this.validationServiceMock.validateSetEventSchema(any())).thenReturn(ValidationResult.ok());
        when(this.validationServiceMock.validateSignatureKey(any())).thenReturn(ValidationResult.ok());
        when(this.keyServiceMock.getPublicSignatureKey(any(),any())).thenReturn(rsaKey);

        // When
        final List<ValidationResult> validationResults = this.underTest.validateEventLog(eventLogPayload, log);

        // Then
        assertFalse(validationResults.isEmpty());
        assertThat(validationResults.get(0).getError().getMessage(), containsString("Destination id '44567b87-6f6a---4f86-9d55-5ff98775d76f' from issuer is no valid uuid"));
    }

    @Test
    void testIssuerDestinationDoesNotMatchSubmission() throws ParseException, JOSEException, IOException {

        // Given
        final RSAKey rsaKey = getRsaKey();
        final JWSHeader header = getDefaultJwsHeader(rsaKey);

        final UUID caseId = UUID.randomUUID();
        final UUID submissionId = UUID.randomUUID();

        final Map<String, Map<String, Object>> events = getEvents(List.of(), ACCEPT);

        final JWTClaimsSet claims = new JWTClaimsSet.Builder()
                .claim(EventClaimFields.CLAIM_SCHEMA, new ApplicationConfig().getSetSchemaWriteVersion())
                .issuer(UUID.randomUUID().toString())
                .issueTime(new Date())
                .jwtID(UUID.randomUUID().toString())
                .subject("submission:" + submissionId)
                .claim(EventClaimFields.CLAIM_TXN, "case:" + caseId)
                .claim(EventClaimFields.CLAIM_EVENTS, events)
                .build();
        final String signedSET = getSignedSET(rsaKey, header, claims);

        final EventLog log = new EventLog();
        log.setEventLogs(List.of(signedSET));

        final EventLogPayload eventLogPayload = EventLogPayload.builder()
                .submissionId(submissionId)
                .destinationId(UUID.randomUUID())
                .caseId(caseId)
                .build();

        when(this.validationServiceMock.validateSetEventSchema(any())).thenReturn(ValidationResult.ok());
        when(this.validationServiceMock.validateSignatureKey(any())).thenReturn(ValidationResult.ok());
        when(this.keyServiceMock.getPublicSignatureKey(any(),any())).thenReturn(rsaKey);

        // When
        final List<ValidationResult> validationResults = this.underTest.validateEventLog(eventLogPayload, log);

        // Then
        assertFalse(validationResults.isEmpty());
        assertThat(validationResults.get(0).getError().getMessage(), containsString("The destination of the submission is not the issuer ('iss' claim must match submission.destinationId)"));
    }

    @Test
    void testIssuerIsNull() throws ParseException, JOSEException, IOException {

        // Given
        final RSAKey rsaKey = getRsaKey();
        final JWSHeader header = getDefaultJwsHeader(rsaKey);

        final UUID caseId = UUID.randomUUID();
        final UUID destinationId = UUID.randomUUID();
        final UUID submissionId = UUID.randomUUID();

        final Map<String, Map<String, Object>> events = getEvents(List.of(), ACCEPT);

        final JWTClaimsSet claims = new JWTClaimsSet.Builder()
                .claim(EventClaimFields.CLAIM_SCHEMA, new ApplicationConfig().getSetSchemaWriteVersion())
                .issuer(null)
                .issueTime(new Date())
                .jwtID(UUID.randomUUID().toString())
                .subject("submission:" + submissionId)
                .claim(EventClaimFields.CLAIM_TXN, "case:" + caseId)
                .claim(EventClaimFields.CLAIM_EVENTS, events)
                .build();
        final String signedSET = getSignedSET(rsaKey, header, claims);

        final EventLog log = new EventLog();
        log.setEventLogs(List.of(signedSET));

        final EventLogPayload eventLogPayload = EventLogPayload.builder()
                .submissionId(submissionId)
                .destinationId(destinationId)
                .caseId(caseId)
                .build();

        when(this.validationServiceMock.validateSetEventSchema(any())).thenReturn(ValidationResult.ok());
        when(this.validationServiceMock.validateSignatureKey(any())).thenReturn(ValidationResult.ok());
        when(this.keyServiceMock.getPublicSignatureKey(any(),any())).thenReturn(rsaKey);

        // When
        final List<ValidationResult> validationResults = this.underTest.validateEventLog(eventLogPayload, log);

        // Then
        assertFalse(validationResults.isEmpty());
        assertThat(validationResults.get(0).getError().getMessage(), containsString("The claim iss is missing in the payload of the SET"));
    }

    @Test
    void testIssuerIsSubmissionService() throws ParseException, JOSEException, IOException {

        // Given
        final RSAKey rsaKey = getRsaKey();
        final JWSHeader header = getDefaultJwsHeader(rsaKey);

        final UUID caseId = UUID.randomUUID();
        final UUID destinationId = UUID.randomUUID();
        final UUID submissionId = UUID.randomUUID();

        final Map<String, Map<String, Object>> events = getEvents(List.of(), ACCEPT);

        final JWTClaimsSet claims = new JWTClaimsSet.Builder()
                .claim(EventClaimFields.CLAIM_SCHEMA, new ApplicationConfig().getSetSchemaWriteVersion())
                .issuer("https://submission-service.fitconnect.de")
                .issueTime(new Date())
                .jwtID(destinationId.toString())
                .subject("submission:" + submissionId)
                .claim(EventClaimFields.CLAIM_TXN, "case:" + caseId)
                .claim(EventClaimFields.CLAIM_EVENTS, events)
                .build();

        final String signedSET = getSignedSET(rsaKey, header, claims);

        final EventLog log = new EventLog();
        log.setEventLogs(List.of(signedSET));

        final EventLogPayload eventLogPayload = EventLogPayload.builder()
                .submissionId(submissionId)
                .destinationId(destinationId)
                .caseId(caseId)
                .build();

        when(this.validationServiceMock.validateSetEventSchema(any())).thenReturn(ValidationResult.ok());
        when(this.validationServiceMock.validateSignatureKey(any())).thenReturn(ValidationResult.ok());
        when(this.keyServiceMock.getPortalSignatureKey(any())).thenReturn(rsaKey);

        // When
        final List<ValidationResult> validationResults = this.underTest.validateEventLog(eventLogPayload, log);

        // Then
        assertTrue(validationResults.isEmpty());
    }

    @Test
    void testInvalidTransaction() throws ParseException, JOSEException, IOException {

        // Given
        final RSAKey rsaKey = getRsaKey();
        final JWSHeader header = getDefaultJwsHeader(rsaKey);

        final UUID caseId = UUID.randomUUID();
        final UUID destinationId = UUID.randomUUID();
        final UUID submissionId = UUID.randomUUID();

        final Map<String, Map<String, Object>> events = getEvents(List.of(), ACCEPT);

        final JWTClaimsSet claims = new JWTClaimsSet.Builder()
                .claim(EventClaimFields.CLAIM_SCHEMA, new ApplicationConfig().getSetSchemaWriteVersion())
                .issuer(destinationId.toString())
                .issueTime(new Date())
                .jwtID(UUID.randomUUID().toString())
                .subject("submission:" + submissionId)
                .claim(EventClaimFields.CLAIM_TXN, "käze:" + caseId)
                .claim(EventClaimFields.CLAIM_EVENTS, events)
                .build();
        final String signedSET = getSignedSET(rsaKey, header, claims);

        final EventLog log = new EventLog();
        log.setEventLogs(List.of(signedSET));

        final EventLogPayload eventLogPayload = EventLogPayload.builder()
                .submissionId(submissionId)
                .destinationId(destinationId)
                .caseId(caseId)
                .build();

        when(this.validationServiceMock.validateSetEventSchema(any())).thenReturn(ValidationResult.ok());
        when(this.validationServiceMock.validateSignatureKey(any())).thenReturn(ValidationResult.ok());
        when(this.keyServiceMock.getPublicSignatureKey(any(), any())).thenReturn(rsaKey);

        // When
        final List<ValidationResult> validationResults = this.underTest.validateEventLog(eventLogPayload, log);

        // Then
        assertFalse(validationResults.isEmpty());
        assertThat(validationResults.get(0).getError().getMessage(), containsString("The provided txn does not match the allowed pattern"));
    }

    @Test
    void testInvalidSubject() throws ParseException, JOSEException, IOException {

        // Given
        final RSAKey rsaKey = getRsaKey();
        final JWSHeader header = getDefaultJwsHeader(rsaKey);

        final UUID caseId = UUID.randomUUID();
        final UUID destinationId = UUID.randomUUID();
        final UUID submissionId = UUID.randomUUID();

        final Map<String, Map<String, Object>> events = getEvents(List.of(), ACCEPT);

        final JWTClaimsSet claims = new JWTClaimsSet.Builder()
                .claim(EventClaimFields.CLAIM_SCHEMA, new ApplicationConfig().getSetSchemaWriteVersion())
                .issuer(destinationId.toString())
                .issueTime(new Date())
                .jwtID(UUID.randomUUID().toString())
                .subject("sub12345")
                .claim(EventClaimFields.CLAIM_TXN, "case:" + caseId)
                .claim(EventClaimFields.CLAIM_EVENTS, events)
                .build();
        final String signedSET = getSignedSET(rsaKey, header, claims);

        final EventLog log = new EventLog();
        log.setEventLogs(List.of(signedSET));

        final EventLogPayload eventLogPayload = EventLogPayload.builder()
                .submissionId(submissionId)
                .destinationId(destinationId)
                .caseId(caseId)
                .build();

        when(this.validationServiceMock.validateSetEventSchema(any())).thenReturn(ValidationResult.ok());
        when(this.validationServiceMock.validateSignatureKey(any())).thenReturn(ValidationResult.ok());
        when(this.keyServiceMock.getPublicSignatureKey(any(),any())).thenReturn(rsaKey);

        // When
        final List<ValidationResult> validationResults = this.underTest.validateEventLog(eventLogPayload, log);

        // Then
        assertFalse(validationResults.isEmpty());
        assertThat(validationResults.get(0).getError().getMessage(), containsString("The provided subject does not match the allowed pattern"));
    }

    @Test
    void testMissingEvent() throws ParseException, JOSEException, IOException {

        // Given
        final RSAKey rsaKey = getRsaKey();
        final JWSHeader header = getDefaultJwsHeader(rsaKey);

        final UUID caseId = UUID.randomUUID();
        final UUID destinationId = UUID.randomUUID();
        final UUID submissionId = UUID.randomUUID();

        final JWTClaimsSet claims = new JWTClaimsSet.Builder()
                .claim(EventClaimFields.CLAIM_SCHEMA, new ApplicationConfig().getSetSchemaWriteVersion())
                .issuer(destinationId.toString())
                .issueTime(new Date())
                .jwtID(UUID.randomUUID().toString())
                .subject("submission:" + submissionId)
                .claim(EventClaimFields.CLAIM_TXN, "case:" + caseId)
                .claim(EventClaimFields.CLAIM_EVENTS, Collections.emptyMap())
                .build();

        final String signedSET = getSignedSET(rsaKey, header, claims);

        final EventLog log = new EventLog();
        log.setEventLogs(List.of(signedSET));

        final EventLogPayload eventLogPayload = EventLogPayload.builder()
                .submissionId(submissionId)
                .destinationId(destinationId)
                .caseId(caseId)
                .build();

        when(this.validationServiceMock.validateSetEventSchema(any())).thenReturn(ValidationResult.ok());
        when(this.validationServiceMock.validateSignatureKey(any())).thenReturn(ValidationResult.ok());
        when(this.keyServiceMock.getPublicSignatureKey(any(),any())).thenReturn(rsaKey);

        // When
        final List<ValidationResult> validationResults = this.underTest.validateEventLog(eventLogPayload, log);

        // Then
        assertFalse(validationResults.isEmpty());
        assertThat(validationResults.get(0).getError().getMessage(), containsString("The events claims has must be exactly one event"));
    }

    @Test
    void testUnsupportedEvent() throws ParseException, JOSEException, IOException {

        // Given
        final RSAKey rsaKey = getRsaKey();
        final JWSHeader header = getDefaultJwsHeader(rsaKey);

        final UUID caseId = UUID.randomUUID();
        final UUID destinationId = UUID.randomUUID();
        final UUID submissionId = UUID.randomUUID();

        final Map<String, Map<String, Object>> unsupportedClientEvent = Map.of("unsupported-event", Map.of("problems", List.of()));

        final JWTClaimsSet claims = new JWTClaimsSet.Builder()
                .claim(EventClaimFields.CLAIM_SCHEMA, new ApplicationConfig().getSetSchemaWriteVersion())
                .issuer(destinationId.toString())
                .issueTime(new Date())
                .jwtID(UUID.randomUUID().toString())
                .subject("submission:" + submissionId)
                .claim(EventClaimFields.CLAIM_TXN, "case:" + caseId)
                .claim(EventClaimFields.CLAIM_EVENTS, unsupportedClientEvent)
                .build();

        final String signedSET = getSignedSET(rsaKey, header, claims);

        final EventLog log = new EventLog();
        log.setEventLogs(List.of(signedSET));

        final EventLogPayload eventLogPayload = EventLogPayload.builder()
                .submissionId(submissionId)
                .destinationId(destinationId)
                .caseId(caseId)
                .build();

        when(this.validationServiceMock.validateSetEventSchema(any())).thenReturn(ValidationResult.ok());
        when(this.validationServiceMock.validateSignatureKey(any())).thenReturn(ValidationResult.ok());
        when(this.keyServiceMock.getPublicSignatureKey(any(),any())).thenReturn(rsaKey);

        // When
        final List<ValidationResult> validationResults = this.underTest.validateEventLog(eventLogPayload, log);

        // Then
        assertFalse(validationResults.isEmpty());
        assertThat(validationResults.get(0).getError().getMessage(), containsString("The provided event is not a valid event supported by this instance"));
    }

    @Test
    void testSignatureKeyUnavailable() throws ParseException, JOSEException, IOException {

        // Given
        final RSAKey rsaKey = getRsaKey();
        final JWSHeader header = getDefaultJwsHeader(rsaKey);

        final UUID caseId = UUID.randomUUID();
        final UUID destinationId = UUID.randomUUID();
        final UUID submissionId = UUID.randomUUID();


        final JWTClaimsSet claims = new JWTClaimsSet.Builder()
                .claim(EventClaimFields.CLAIM_SCHEMA, new ApplicationConfig().getSetSchemaWriteVersion())
                .issuer(destinationId.toString())
                .issueTime(new Date())
                .jwtID(UUID.randomUUID().toString())
                .subject("submission:" + submissionId)
                .claim(EventClaimFields.CLAIM_TXN, "case:" + caseId)
                .claim(EventClaimFields.CLAIM_EVENTS, getEvents(List.of(), ACCEPT))
                .build();

        final String signedSET = getSignedSET(rsaKey, header, claims);

        final EventLog log = new EventLog();
        log.setEventLogs(List.of(signedSET));

        final EventLogPayload eventLogPayload = EventLogPayload.builder()
                .submissionId(submissionId)
                .destinationId(destinationId)
                .caseId(caseId)
                .build();

        when(this.validationServiceMock.validateSetEventSchema(any())).thenReturn(ValidationResult.ok());
        when(this.validationServiceMock.validateSignatureKey(any())).thenReturn(ValidationResult.ok());
        when(this.keyServiceMock.getPublicSignatureKey(any(),any())).thenReturn(null);

        // When
        final List<ValidationResult> validationResults = this.underTest.validateEventLog(eventLogPayload, log);

        // Then
        assertFalse(validationResults.isEmpty());
        assertThat(validationResults.get(0).getError().getMessage(), containsString("The signature key cannot be validated, it must not be null"));
    }

    private static JWTClaimsSet getDefaultClaims(final UUID submissionId, final UUID caseId, final UUID destinationId, final Map<String, Map<String, Object>> events) {
        return new JWTClaimsSet.Builder()
                .claim(EventClaimFields.CLAIM_SCHEMA, new ApplicationConfig().getSetSchemaWriteVersion())
                .issuer(destinationId.toString())
                .issueTime(new Date())
                .jwtID(UUID.randomUUID().toString())
                .subject("submission:" + submissionId)
                .claim(EventClaimFields.CLAIM_TXN, "case:" + caseId)
                .claim(EventClaimFields.CLAIM_EVENTS, events)
                .build();
    }

    private static Map<String, Map<String, Object>> getEvents(final List<Problem> problems, final Event event) {
        return Map.of(event.getSchema(), Map.of("problems", problems));
    }

    private static JWSHeader getDefaultJwsHeader(final RSAKey rsaKey) throws ParseException {
        return JWSHeader.parse(Map.of(
                EventHeaderFields.TYPE, "secevent+jwt",
                EventHeaderFields.KEY_ID, rsaKey.getKeyID(),
                EventHeaderFields.ALGORITHM, "PS512"
        ));
    }
    private static String getSignedSET(final RSAKey rsaKey, final JWSHeader header, final JWTClaimsSet claims) throws JOSEException {
        final SignedJWT signedJWT = new SignedJWT(header, claims);
        signedJWT.sign( new RSASSASigner(rsaKey));
        return signedJWT.serialize();
    }

    private RSAKey getRsaKey() throws IOException, ParseException {
        return RSAKey.parse(getResourceAsString("private_test_verify_key.json"));
    }

    private String getResourceAsString(final String name) throws IOException {
        final ClassLoader classLoader = getClass().getClassLoader();
        return Files.readString(new File(classLoader.getResource(name).getFile()).toPath());
    }
}