package dev.fitko.fitconnect.core.events;

import dev.fitko.fitconnect.api.domain.model.event.EventLogPayload;
import dev.fitko.fitconnect.api.domain.validation.ValidationResult;
import org.junit.jupiter.api.Test;

import java.util.List;
import java.util.UUID;

import static org.hamcrest.MatcherAssert.assertThat;
import static org.hamcrest.Matchers.is;
import static org.junit.jupiter.api.Assertions.assertEquals;
import static org.junit.jupiter.api.Assertions.assertTrue;

class ValidationContextTest {

    @Test
    void addPositiveTestResult() {

        // Given
        final ValidationContext underTest = new ValidationContext(new EventLogPayload());

        // When
        underTest.addResult(1+2 == 3, "1+2 does not equal 3");
        final List<ValidationResult> results = underTest.getValidationResults();

        // Then
        assertTrue(results.isEmpty());
    }

    @Test
    void addNegativeTestResult() {

        // Given
        final ValidationContext underTest = new ValidationContext(new EventLogPayload());

        // When
        underTest.addResult(1+2 == 2, "1+2 does not equal 3");
        final List<ValidationResult> results = underTest.getValidationResults();

        // Then
        assertEquals(1, results.size());
        assertTrue(results.get(0).hasError());
        assertThat(results.get(0).getError().getMessage(), is("1+2 does not equal 3"));
    }

    @Test
    void testAddPositiveValidationResult() {

        // Given
        final ValidationContext underTest = new ValidationContext(new EventLogPayload());

        // When
        underTest.addResult(ValidationResult.ok());
        final List<ValidationResult> results = underTest.getValidationResults();

        // Then
        assertTrue(results.isEmpty());
    }

    @Test
    void testAddNegativeValidationResult() {

        // Given
        final ValidationContext underTest = new ValidationContext(new EventLogPayload());

        // When
        underTest.addResult(ValidationResult.error(new Exception("Test failed")));
        final List<ValidationResult> results = underTest.getValidationResults();

        // Then
        assertTrue(results.size() == 1);
        assertTrue(results.get(0).hasError());
        assertThat(results.get(0).getError().getMessage(), is("Test failed"));
    }

    @Test
    void testAddError() {

        // Given
        final ValidationContext underTest = new ValidationContext(new EventLogPayload());

        // When
        underTest.addError("Test failed");
        final List<ValidationResult> results = underTest.getValidationResults();

        // Then
        assertTrue(results.size() == 1);
        assertTrue(results.get(0).hasError());
        assertThat(results.get(0).getError().getMessage(), is("Test failed"));
    }

    @Test
    void testGetDestinationId() {

        // Given
        final EventLogPayload eventLogPayload = new EventLogPayload();
        eventLogPayload.setDestinationId(UUID.randomUUID());

        //When
        final ValidationContext underTest = new ValidationContext(eventLogPayload);

        // Then
        assertThat(underTest.getDestinationId(), is(eventLogPayload.getDestinationId()));
    }

    @Test
    void testGetCaseId() {

        // Given
        final EventLogPayload eventLogPayload = new EventLogPayload();
        eventLogPayload.setCaseId(UUID.randomUUID());

        //When
        final ValidationContext underTest = new ValidationContext(eventLogPayload);

        // Then
        assertThat(underTest.getCaseId(), is(eventLogPayload.getCaseId()));
    }

    @Test
    void getEventLogPayload() {

        // Given
        final EventLogPayload expectedLog = new EventLogPayload();
        expectedLog.setCaseId(UUID.randomUUID());
        expectedLog.setEncryptedData("encrypted data");

        final ValidationContext underTest = new ValidationContext(expectedLog);

        //When
        final EventLogPayload log = underTest.getEventLogPayload();

        // Then
        assertThat(log, is(expectedLog));
    }
}